@@ -1,8 +1,9 @@
 # This relies on each of the submodules having an __all__ variable.
 
-__version__ = '0.13.1'
+__version__ = '0.14.1'
 
 
+from . import hdrs  # noqa
 from .protocol import *  # noqa
 from .connector import *  # noqa
 from .client import *  # noqa
@@ -10,6 +11,7 @@ from .errors import *  # noqa
 from .helpers import *  # noqa
 from .parsers import *  # noqa
 from .streams import *  # noqa
+from .multidict import *  # noqa
 
 
 __all__ = (client.__all__ +
@@ -19,4 +21,5 @@ __all__ = (client.__all__ +
            protocol.__all__ +
            connector.__all__ +
            streams.__all__ +
-           ['__version__'])
+           multidict.__all__ +
+           ['hdrs', '__version__'])
