@@ -16,10 +16,10 @@ except ImportError:  # pragma: no cover
     chardet = None
 
 import aiohttp
-from . import helpers, streams
+from . import hdrs, helpers, streams
 from .log import client_logger
 from .streams import EOF_MARKER, FlowControlStreamReader
-from .multidict import CaseInsensitiveMultiDict, MultiDict, MutableMultiDict
+from .multidict import CIMultiDictProxy, MultiDictProxy, MultiDict
 
 HTTP_PORT = 80
 HTTPS_PORT = 443
@@ -58,8 +58,8 @@ def request(method, url, *,
     :param dict cookies: (optional) Dict object to send with the request
     :param auth: (optional) BasicAuth named tuple represent HTTP Basic Auth
     :type auth: aiohttp.helpers.BasicAuth
-    :param bool allow_redirects: (optional) Set to True if POST/PUT/DELETE
-       redirect following is allowed.
+    :param bool allow_redirects: (optional) If set to False, do not follow
+      redirects
     :param version: Request http version.
     :type version: aiohttp.protocol.HttpVersion
     :param bool compress: Set to True if request has to be compressed
@@ -130,7 +130,8 @@ def request(method, url, *,
                 data = None
             cookies = resp.cookies
 
-            r_url = resp.headers.get('LOCATION') or resp.headers.get('URI')
+            r_url = (resp.headers.get(hdrs.LOCATION) or
+                     resp.headers.get(hdrs.URI))
 
             scheme = urllib.parse.urlsplit(r_url)[0]
             if scheme not in ('http', 'https', ''):
@@ -156,8 +157,8 @@ class ClientRequest:
     ALL_METHODS = GET_METHODS.union(POST_METHODS)
 
     DEFAULT_HEADERS = {
-        'ACCEPT': '*/*',
-        'ACCEPT-ENCODING': 'gzip, deflate',
+        hdrs.ACCEPT: '*/*',
+        hdrs.ACCEPT_ENCODING: 'gzip, deflate',
     }
 
     body = b''
@@ -275,8 +276,8 @@ class ClientRequest:
 
         if isinstance(params, dict):
             params = list(params.items())
-        elif isinstance(params, MultiDict):
-            params = list(params.items(getall=True))
+        elif isinstance(params, (MultiDictProxy, MultiDict)):
+            params = list(params.items())
 
         if params:
             params = urllib.parse.urlencode(params)
@@ -286,16 +287,16 @@ class ClientRequest:
                 query = params
 
         self.path = urllib.parse.urlunsplit(
-            ('', '', urllib.parse.quote(path, safe='/%'), query, fragment))
+            ('', '', urllib.parse.quote(path, safe='/%:'), query, fragment))
 
     def update_headers(self, headers):
         """Update request headers."""
-        self.headers = MutableMultiDict()
+        self.headers = MultiDict()
         if headers:
             if isinstance(headers, dict):
                 headers = headers.items()
-            elif isinstance(headers, MultiDict):
-                headers = headers.items(getall=True)
+            elif isinstance(headers, (MultiDictProxy, MultiDict)):
+                headers = headers.items()
 
             for key, value in headers:
                 self.headers.add(key.upper(), value)
@@ -305,8 +306,8 @@ class ClientRequest:
                 self.headers[hdr] = val
 
         # add host
-        if 'HOST' not in self.headers:
-            self.headers['HOST'] = self.netloc
+        if hdrs.HOST not in self.headers:
+            self.headers[hdrs.HOST] = self.netloc
 
     def update_cookies(self, cookies):
         """Update request cookies header."""
@@ -314,9 +315,9 @@ class ClientRequest:
             return
 
         c = http.cookies.SimpleCookie()
-        if 'COOKIE' in self.headers:
-            c.load(self.headers.get('COOKIE', ''))
-            del self.headers['COOKIE']
+        if hdrs.COOKIE in self.headers:
+            c.load(self.headers.get(hdrs.COOKIE, ''))
+            del self.headers[hdrs.COOKIE]
 
         if isinstance(cookies, dict):
             cookies = cookies.items()
@@ -328,18 +329,18 @@ class ClientRequest:
             else:
                 c[name] = value
 
-        self.headers['COOKIE'] = c.output(header='', sep=';').strip()
+        self.headers[hdrs.COOKIE] = c.output(header='', sep=';').strip()
 
     def update_content_encoding(self):
         """Set request content encoding."""
-        enc = self.headers.get('CONTENT-ENCODING', '').lower()
+        enc = self.headers.get(hdrs.CONTENT_ENCODING, '').lower()
         if enc:
             self.compress = enc
             self.chunked = True  # enable chunked, no need to deal with length
         elif self.compress:
             if not isinstance(self.compress, str):
                 self.compress = 'deflate'
-            self.headers['CONTENT-ENCODING'] = self.compress
+            self.headers[hdrs.CONTENT_ENCODING] = self.compress
             self.chunked = True  # enable chunked, no need to deal with length
 
     def update_auth(self, auth):
@@ -354,7 +355,7 @@ class ClientRequest:
                 'BasicAuth() tuple is required instead ', DeprecationWarning)
             auth = helpers.BasicAuth(*auth)
 
-        self.headers['AUTHORIZATION'] = auth.encode()
+        self.headers[hdrs.AUTHORIZATION] = auth.encode()
 
     def update_body_from_data(self, data):
         if not data:
@@ -365,17 +366,18 @@ class ClientRequest:
 
         if isinstance(data, (bytes, bytearray)):
             self.body = data
-            if 'CONTENT-TYPE' not in self.headers:
-                self.headers['CONTENT-TYPE'] = 'application/octet-stream'
-            if 'CONTENT-LENGTH' not in self.headers and not self.chunked:
-                self.headers['CONTENT-LENGTH'] = str(len(self.body))
+            if hdrs.CONTENT_TYPE not in self.headers:
+                self.headers[hdrs.CONTENT_TYPE] = 'application/octet-stream'
+            if hdrs.CONTENT_LENGTH not in self.headers and not self.chunked:
+                self.headers[hdrs.CONTENT_LENGTH] = str(len(self.body))
 
         elif isinstance(data, (asyncio.StreamReader, streams.DataQueue)):
             self.body = data
 
         elif asyncio.iscoroutine(data):
             self.body = data
-            if 'CONTENT-LENGTH' not in self.headers and self.chunked is None:
+            if (hdrs.CONTENT_LENGTH not in self.headers and
+                    self.chunked is None):
                 self.chunked = True
 
         elif isinstance(data, io.IOBase):
@@ -387,10 +389,11 @@ class ClientRequest:
                 if data.mode == 'r':
                     raise ValueError('file {!r} should be open in binary mode'
                                      ''.format(data))
-            if 'CONTENT-TYPE' not in self.headers and hasattr(data, 'name'):
+            if (hdrs.CONTENT_TYPE not in self.headers and
+                    hasattr(data, 'name')):
                 mime = mimetypes.guess_type(data.name)[0]
                 mime = 'application/octet-stream' if mime is None else mime
-                self.headers['CONTENT-TYPE'] = mime
+                self.headers[hdrs.CONTENT_TYPE] = mime
 
         else:
             if not isinstance(data, helpers.FormData):
@@ -398,24 +401,25 @@ class ClientRequest:
 
             self.body = data(self.encoding)
 
-            if 'CONTENT-TYPE' not in self.headers:
-                self.headers['CONTENT-TYPE'] = data.content_type
+            if hdrs.CONTENT_TYPE not in self.headers:
+                self.headers[hdrs.CONTENT_TYPE] = data.content_type
 
             if data.is_multipart:
                 self.chunked = self.chunked or 8192
             else:
-                if 'CONTENT-LENGTH' not in self.headers and not self.chunked:
-                    self.headers['CONTENT-LENGTH'] = str(len(self.body))
+                if (hdrs.CONTENT_LENGTH not in self.headers and
+                        not self.chunked):
+                    self.headers[hdrs.CONTENT_LENGTH] = str(len(self.body))
 
     def update_transfer_encoding(self):
         """Analyze transfer-encoding header."""
-        te = self.headers.get('TRANSFER-ENCODING', '').lower()
+        te = self.headers.get(hdrs.TRANSFER_ENCODING, '').lower()
 
         if self.chunked:
-            if 'CONTENT-LENGTH' in self.headers:
-                del self.headers['CONTENT-LENGTH']
+            if hdrs.CONTENT_LENGTH in self.headers:
+                del self.headers[hdrs.CONTENT_LENGTH]
             if 'chunked' not in te:
-                self.headers['TRANSFER-ENCODING'] = 'chunked'
+                self.headers[hdrs.TRANSFER_ENCODING] = 'chunked'
 
             self.chunked = self.chunked if type(self.chunked) is int else 8192
         else:
@@ -423,13 +427,13 @@ class ClientRequest:
                 self.chunked = 8192
             else:
                 self.chunked = None
-                if 'CONTENT-LENGTH' not in self.headers:
-                    self.headers['CONTENT-LENGTH'] = str(len(self.body))
+                if hdrs.CONTENT_LENGTH not in self.headers:
+                    self.headers[hdrs.CONTENT_LENGTH] = str(len(self.body))
 
     def update_expect_continue(self, expect=False):
         if expect:
-            self.headers['EXPECT'] = '100-continue'
-        elif self.headers.get('EXPECT', '').lower() == '100-continue':
+            self.headers[hdrs.EXPECT] = '100-continue'
+        elif self.headers.get(hdrs.EXPECT, '').lower() == '100-continue':
             expect = True
 
         if expect:
@@ -535,12 +539,13 @@ class ClientRequest:
             request.add_compression_filter(self.compress)
 
         if self.chunked is not None:
+            request.enable_chunked_encoding()
             request.add_chunking_filter(self.chunked)
 
         request.add_headers(
             *((k, v)
               for k, v in ((k, value)
-                           for k, value in self.headers.items(getall=True))))
+                           for k, value in self.headers.items())))
         request.send_headers()
 
         self._writer = asyncio.async(
@@ -642,8 +647,7 @@ class ClientResponse:
         self.reason = self.message.reason
 
         # headers
-        self.headers = CaseInsensitiveMultiDict(
-            self.message.headers.items(getall=True))
+        self.headers = CIMultiDictProxy(self.message.headers)
 
         # payload
         response_with_body = self.method.lower() != 'head'
@@ -655,8 +659,8 @@ class ClientResponse:
 
         # cookies
         self.cookies = http.cookies.SimpleCookie()
-        if 'SET-COOKIE' in self.headers:
-            for hdr in self.headers.getall('SET-COOKIE'):
+        if hdrs.SET_COOKIE in self.headers:
+            for hdr in self.headers.getall(hdrs.SET_COOKIE):
                 try:
                     self.cookies.load(hdr)
                 except http.cookies.CookieError as exc:
@@ -731,14 +735,14 @@ class ClientResponse:
         return (yield from self.read(decode))
 
     def _get_encoding(self, encoding):
-        ctype = self.headers.get('CONTENT-TYPE', '').lower()
+        ctype = self.headers.get(hdrs.CONTENT_TYPE, '').lower()
         mtype, stype, _, params = helpers.parse_mimetype(ctype)
 
         if not encoding:
             encoding = params.get('charset')
             if not encoding and chardet:
                 encoding = chardet.detect(self._content)['encoding']
-            if not encoding:  # pragma: no cover
+            if not encoding:
                 encoding = 'utf-8'
 
         return encoding
@@ -757,7 +761,7 @@ class ClientResponse:
         if self._content is None:
             yield from self.read()
 
-        ctype = self.headers.get('CONTENT-TYPE', '').lower()
+        ctype = self.headers.get(hdrs.CONTENT_TYPE, '').lower()
         if 'json' not in ctype:
             client_logger.warning(
                 'Attempt to decode JSON with unexpected mimetype: %s', ctype)
