@@ -120,6 +120,11 @@ class StreamParser(asyncio.streams.StreamReader):
         return self._exception
 
     def set_exception(self, exc):
+        if isinstance(exc, ConnectionError):
+            exc, old_exc = self._eof_exc_class(), exc
+            exc.__cause__ = old_exc
+            exc.__context__ = old_exc
+
         self._exception = exc
 
         if self._output is not None:
@@ -249,7 +254,7 @@ class StreamProtocol(asyncio.streams.FlowControlMixin, asyncio.Protocol):
             transport, self, self.reader, self._loop)
 
     def connection_lost(self, exc):
-        self.transport = None
+        self.transport = self.writer = None
 
         if exc is None:
             self.reader.feed_eof()
