@@ -17,9 +17,9 @@ import zlib
 from wsgiref.handlers import format_date_time
 
 import aiohttp
-from aiohttp import errors
-from aiohttp import multidict
-from aiohttp.log import internal_logger
+from . import errors, hdrs
+from .multidict import CIMultiDict
+from .log import internal_logger
 
 ASCIISET = set(string.printable)
 METHRE = re.compile('[A-Z0-9$-_.]+')
@@ -63,7 +63,7 @@ class HttpParser:
         """
         close_conn = None
         encoding = None
-        headers = []
+        headers = CIMultiDict()
 
         lines_idx = 1
         line = lines[1]
@@ -110,20 +110,20 @@ class HttpParser:
             value = value.strip()
 
             # keep-alive and encoding
-            if name == 'CONNECTION':
+            if name == hdrs.CONNECTION:
                 v = value.lower()
                 if v == 'close':
                     close_conn = True
                 elif v == 'keep-alive':
                     close_conn = False
-            elif name == 'CONTENT-ENCODING':
+            elif name == hdrs.CONTENT_ENCODING:
                 enc = value.lower()
                 if enc in ('gzip', 'deflate'):
                     encoding = enc
 
-            headers.append((name, value))
+            headers.add(name, value)
 
-        return multidict.MultiDict(headers), close_conn, encoding
+        return headers, close_conn, encoding
 
 
 class HttpPrefixParser:
@@ -265,8 +265,8 @@ class HttpPayloadParser:
 
     def __call__(self, out, buf):
         # payload params
-        length = self.message.headers.get('CONTENT-LENGTH', self.length)
-        if 'SEC-WEBSOCKET-KEY1' in self.message.headers:
+        length = self.message.headers.get(hdrs.CONTENT_LENGTH, self.length)
+        if hdrs.SEC_WEBSOCKET_KEY1 in self.message.headers:
             length = 8
 
         # payload decompression wrapper
@@ -278,17 +278,18 @@ class HttpPayloadParser:
             # don't parse payload if it's not expected to be received
             pass
 
-        elif 'chunked' in self.message.headers.get('TRANSFER-ENCODING', ''):
+        elif 'chunked' in self.message.headers.get(
+                hdrs.TRANSFER_ENCODING, ''):
             yield from self.parse_chunked_payload(out, buf)
 
         elif length is not None:
             try:
                 length = int(length)
             except ValueError:
-                raise errors.InvalidHeader('CONTENT-LENGTH') from None
+                raise errors.InvalidHeader(hdrs.CONTENT_LENGTH) from None
 
             if length < 0:
-                raise errors.InvalidHeader('CONTENT-LENGTH')
+                raise errors.InvalidHeader(hdrs.CONTENT_LENGTH)
             elif length > 0:
                 yield from self.parse_length_payload(out, buf, length)
         else:
@@ -423,7 +424,8 @@ def wrap_payload_filter(func):
     return wrapper
 
 
-def filter_pipe(filter, filter2):
+def filter_pipe(filter, filter2, *,
+                EOF_MARKER=EOF_MARKER, EOL_MARKER=EOL_MARKER):
     """Creates pipe between two filters.
 
     filter_pipe() feeds first filter with incoming data and then
@@ -516,13 +518,12 @@ class HttpMessage:
     status_line = b''
     upgrade = False  # Connection: UPGRADE
     websocket = False  # Upgrade: WEBSOCKET
+    has_chunked_hdr = False  # Transfer-encoding: chunked
 
     # subclass can enable auto sending headers with write() call,
     # this is useful for wsgi's start_response implementation.
     _send_headers = False
 
-    _has_user_agent = False
-
     def __init__(self, transport, version, close):
         self.transport = transport
         self.version = version
@@ -536,7 +537,7 @@ class HttpMessage:
 
         self.chunked = False
         self.length = None
-        self.headers = multidict.CaseInsensitiveMutableMultiDict()
+        self.headers = CIMultiDict()
         self.headers_sent = False
         self.output_length = 0
         self._output_size = 0
@@ -545,7 +546,7 @@ class HttpMessage:
         self.closing = True
         self.keepalive = False
 
-    def force_chunked(self):
+    def enable_chunked_encoding(self):
         self.chunked = True
 
     def keep_alive(self):
@@ -572,10 +573,13 @@ class HttpMessage:
         name = name.strip().upper()
         value = value.strip()
 
-        if name == 'CONTENT-LENGTH':
+        if name == hdrs.CONTENT_LENGTH:
             self.length = int(value)
 
-        if name == 'CONNECTION':
+        if name == hdrs.TRANSFER_ENCODING:
+            self.has_chunked_hdr = value.lower().strip() == 'chunked'
+
+        if name == hdrs.CONNECTION:
             val = value.lower()
             # handle websocket
             if 'upgrade' in val:
@@ -586,18 +590,12 @@ class HttpMessage:
             elif 'keep-alive' in val and self.version >= HttpVersion11:
                 self.keepalive = True
 
-        elif name == 'UPGRADE':
+        elif name == hdrs.UPGRADE:
             if 'websocket' in value.lower():
                 self.websocket = True
                 self.headers[name] = value
 
-        elif name == 'TRANSFER-ENCODING' and not self.chunked:
-            self.chunked = value.lower().strip() == 'chunked'
-
         elif name not in self.HOP_HEADERS:
-            if name == 'USER-AGENT':
-                self._has_user_agent = True
-
             # ignore hop-by-hop headers
             self.headers.add(name, value)
 
@@ -615,12 +613,11 @@ class HttpMessage:
         assert not self.headers_sent, 'headers have been sent already'
         self.headers_sent = True
 
-        if (self.chunked is True) or (
-                self.length is None and
-                self.version >= HttpVersion11 and
-                self.status not in (304, 204)):
-            self.chunked = True
+        if self.chunked or (self.length is None and
+                            self.version >= HttpVersion11 and
+                            self.status not in (304, 204)):
             self.writer = self._write_chunked_payload()
+            self.headers[hdrs.TRANSFER_ENCODING] = 'chunked'
 
         elif self.length is not None:
             self.writer = self._write_length_payload(self.length)
@@ -633,15 +630,15 @@ class HttpMessage:
         self._add_default_headers()
 
         # status + headers
-        hdrs = ''.join(itertools.chain(
+        headers = ''.join(itertools.chain(
             (self.status_line,),
             *((k, ': ', v, '\r\n')
               for k, v in ((k, value)
-                           for k, value in self.headers.items(getall=True)))))
-        hdrs = hdrs.encode('utf-8') + b'\r\n'
+                           for k, value in self.headers.items()))))
+        headers = headers.encode('utf-8') + b'\r\n'
 
-        self.output_length += len(hdrs)
-        self.transport.write(hdrs)
+        self.output_length += len(headers)
+        self.transport.write(headers)
 
     def _add_default_headers(self):
         # set the connection header
@@ -652,12 +649,9 @@ class HttpMessage:
         else:
             connection = 'close'
 
-        if self.chunked:
-            self.headers['TRANSFER-ENCODING'] = 'chunked'
+        self.headers[hdrs.CONNECTION] = connection
 
-        self.headers['CONNECTION'] = connection
-
-    def write(self, chunk):
+    def write(self, chunk, *, EOF_MARKER=EOF_MARKER, EOL_MARKER=EOL_MARKER):
         """Writes chunk of data to a stream by using different writers.
 
         writer uses filter to modify chunk of data.
@@ -748,7 +742,8 @@ class HttpMessage:
             self.output_length += len(chunk)
 
     @wrap_payload_filter
-    def add_chunking_filter(self, chunk_size=16*1024):
+    def add_chunking_filter(self, chunk_size=16*1024, *,
+                            EOF_MARKER=EOF_MARKER, EOL_MARKER=EOL_MARKER):
         """Split incoming stream into chunks."""
         buf = bytearray()
         chunk = yield
@@ -771,7 +766,8 @@ class HttpMessage:
                 chunk = yield EOL_MARKER
 
     @wrap_payload_filter
-    def add_compression_filter(self, encoding='deflate'):
+    def add_compression_filter(self, encoding='deflate', *,
+                               EOF_MARKER=EOF_MARKER, EOL_MARKER=EOL_MARKER):
         """Compress incoming stream with deflate or gzip encoding."""
         zlib_mode = (16 + zlib.MAX_WBITS
                      if encoding == 'gzip' else -zlib.MAX_WBITS)
@@ -796,18 +792,7 @@ class Response(HttpMessage):
     http version, (1, 0) stands for HTTP/1.0 and (1, 1) is for HTTP/1.1
     """
 
-    HOP_HEADERS = {
-        'CONNECTION',
-        'KEEP-ALIVE',
-        'PROXY-AUTHENTICATE',
-        'PROXY-AUTHORIZATION',
-        'TE',
-        'TRAILERS',
-        'TRANSFER-ENCODING',
-        'UPGRADE',
-        'SERVER',
-        'DATE',
-    }
+    HOP_HEADERS = ()
 
     @staticmethod
     def calc_reason(status):
@@ -833,8 +818,10 @@ class Response(HttpMessage):
     def _add_default_headers(self):
         super()._add_default_headers()
 
-        self.headers.extend((('DATE', format_date_time(None)),
-                             ('SERVER', self.SERVER_SOFTWARE),))
+        if hdrs.DATE not in self.headers:
+            # format_date_time(None) is quite expensive
+            self.headers.setdefault(hdrs.DATE, format_date_time(None))
+        self.headers.setdefault(hdrs.SERVER, self.SERVER_SOFTWARE)
 
 
 class Request(HttpMessage):
@@ -853,5 +840,4 @@ class Request(HttpMessage):
     def _add_default_headers(self):
         super()._add_default_headers()
 
-        if not self._has_user_agent:
-            self.headers['USER-AGENT'] = self.SERVER_SOFTWARE
+        self.headers.setdefault(hdrs.USER_AGENT, self.SERVER_SOFTWARE)
