@@ -75,7 +75,7 @@ class ServerHttpProtocol(aiohttp.StreamProtocol):
     _request_parser = aiohttp.HttpRequestParser()  # default request parser
 
     def __init__(self, *, loop=None,
-                 keep_alive=None,
+                 keep_alive=75,  # NGINX default value is 75 secs
                  timeout=15,
                  tcp_keepalive=True,
                  allowed_methods=(),
@@ -239,7 +239,7 @@ class ServerHttpProtocol(aiohttp.StreamProtocol):
                         isinstance(handler, asyncio.Future)):
                     yield from handler
 
-            except (ConnectionError, asyncio.CancelledError,
+            except (asyncio.CancelledError,
                     errors.ClientDisconnectedError):
                 self.log_debug('Ignored premature client disconnection.')
                 break
