@@ -11,20 +11,25 @@ import re
 import os
 
 from urllib.parse import urlsplit, parse_qsl, urlencode, unquote
+from types import MappingProxyType
 
+from . import hdrs
 from .abc import AbstractRouter, AbstractMatchInfo
 from .helpers import reify
 from .log import web_logger
-from .multidict import (CaseInsensitiveMultiDict,
-                        CaseInsensitiveMutableMultiDict,
-                        MultiDict,
-                        MutableMultiDict)
+from .multidict import (CIMultiDictProxy,
+                        CIMultiDict,
+                        MultiDictProxy,
+                        MultiDict)
 from .protocol import Response as ResponseImpl, HttpVersion, HttpVersion11
 from .server import ServerHttpProtocol
 from .streams import EOF_MARKER
+from .websocket import do_handshake, MSG_BINARY, MSG_CLOSE, MSG_PING, MSG_TEXT
+from .errors import HttpProcessingError, WSClientDisconnectedError
 
 
 __all__ = [
+    'WSClientDisconnectedError',
     'Application',
     'HttpVersion',
     'RequestHandler',
@@ -32,6 +37,7 @@ __all__ = [
     'Request',
     'StreamResponse',
     'Response',
+    'WebSocketResponse',
     'UrlDispatcher',
     'UrlMappingMatchInfo',
     'HTTPException',
@@ -100,25 +106,25 @@ class HeadersMixin:
             self._content_type, self._content_dict = cgi.parse_header(raw)
 
     @property
-    def content_type(self):
+    def content_type(self, _CONTENT_TYPE=hdrs.CONTENT_TYPE):
         """The value of content part for Content-Type HTTP header."""
-        raw = self.headers.get('Content-Type')
+        raw = self.headers.get(_CONTENT_TYPE)
         if self._stored_content_type != raw:
             self._parse_content_type(raw)
         return self._content_type
 
     @property
-    def charset(self):
+    def charset(self, _CONTENT_TYPE=hdrs.CONTENT_TYPE):
         """The value of charset part for Content-Type HTTP header."""
-        raw = self.headers.get('Content-Type')
+        raw = self.headers.get(_CONTENT_TYPE)
         if self._stored_content_type != raw:
             self._parse_content_type(raw)
         return self._content_dict.get('charset')
 
     @property
-    def content_length(self):
+    def content_length(self, _CONTENT_LENGTH=hdrs.CONTENT_LENGTH):
         """The value of Content-Length HTTP header."""
-        l = self.headers.get('Content-Length')
+        l = self.headers.get(_CONTENT_LENGTH)
         if l is None:
             return None
         else:
@@ -133,31 +139,31 @@ FileField = collections.namedtuple('Field', 'name filename file content_type')
 ############################################################
 
 
-class Request(HeadersMixin):
+class Request(dict, HeadersMixin):
 
-    def __init__(self, app, message, payload, transport, writer,
-                 keep_alive_timeout):
+    def __init__(self, app, message, payload, transport, reader, writer, *,
+                 _HOST=hdrs.HOST):
         self._app = app
         self._version = message.version
         self._transport = transport
+        self._reader = reader
         self._writer = writer
         self._method = message.method
-        self._host = message.headers.get('HOST')
+        self._host = message.headers.get(_HOST)
         self._path_qs = message.path
         res = urlsplit(message.path)
         self._path = unquote(res.path)
         self._query_string = res.query
         self._post = None
         self._post_files_cache = None
-        self._headers = CaseInsensitiveMultiDict._from_uppercase_multidict(
-            message.headers)
+        self._headers = CIMultiDictProxy(message.headers)
 
         if self._version < HttpVersion11:
             self._keep_alive = False
         elif message.should_close:
             self._keep_alive = False
         else:
-            self._keep_alive = bool(keep_alive_timeout)
+            self._keep_alive = True
 
         # matchdict, route_name, handler
         # or information about traversal lookup
@@ -220,7 +226,7 @@ class Request(HeadersMixin):
 
         Lazy property.
         """
-        return MultiDict(parse_qsl(self._query_string))
+        return MultiDictProxy(MultiDict(parse_qsl(self._query_string)))
 
     @reify
     def POST(self):
@@ -267,10 +273,10 @@ class Request(HeadersMixin):
         A read-only dictionary-like object.
         """
         if self._cookies is None:
-            raw = self.headers.get('COOKIE', '')
+            raw = self.headers.get(hdrs.COOKIE, '')
             parsed = http.cookies.SimpleCookie(raw)
-            self._cookies = MultiDict({key: val.value
-                                       for key, val in parsed.items()})
+            self._cookies = MappingProxyType(
+                {key: val.value for key, val in parsed.items()})
         return self._cookies
 
     @property
@@ -321,14 +327,14 @@ class Request(HeadersMixin):
         if self._post is not None:
             return self._post
         if self.method not in ('POST', 'PUT', 'PATCH'):
-            self._post = MultiDict()
+            self._post = MultiDictProxy(MultiDict())
             return self._post
 
         content_type = self.content_type
         if (content_type not in ('',
                                  'application/x-www-form-urlencoded',
                                  'multipart/form-data')):
-            self._post = MultiDict()
+            self._post = MultiDictProxy(MultiDict())
             return self._post
 
         body = yield from self.read()
@@ -337,7 +343,7 @@ class Request(HeadersMixin):
         environ = {'REQUEST_METHOD': self.method,
                    'CONTENT_LENGTH': str(len(body)),
                    'QUERY_STRING': '',
-                   'CONTENT_TYPE': self.headers.get('CONTENT-TYPE')}
+                   'CONTENT_TYPE': self.headers.get(hdrs.CONTENT_TYPE)}
 
         fs = cgi.FieldStorage(fp=io.BytesIO(body),
                               environ=environ,
@@ -349,7 +355,7 @@ class Request(HeadersMixin):
             'quoted-printable': binascii.a2b_qp
         }
 
-        out = MutableMultiDict()
+        out = MultiDict()
         for field in fs.list or ():
             transfer_encoding = field.headers.get(
                 'Content-Transfer-Encoding', None)
@@ -371,20 +377,9 @@ class Request(HeadersMixin):
                         transfer_encoding](value)
                 out.add(field.name, value)
 
-        self._post = MultiDict(out.items(getall=True))
+        self._post = MultiDictProxy(out)
         return self._post
 
-    # @asyncio.coroutine
-    # def start_websocket(self):
-    #     """Upgrade connection to websocket.
-
-    #     Returns (reader, writer) pair.
-    #     """
-
-    #     upgrade = 'websocket' in message.headers.get('UPGRADE', '').lower()
-    #     if not upgrade:
-    #         pass
-
 
 ############################################################
 # HTTP Response classes
@@ -396,7 +391,11 @@ class StreamResponse(HeadersMixin):
     def __init__(self, *, status=200, reason=None):
         self._body = None
         self._keep_alive = None
-        self._headers = CaseInsensitiveMutableMultiDict()
+        self._chunked = False
+        self._chunk_size = None
+        self._compression = False
+        self._compression_force = False
+        self._headers = CIMultiDict()
         self._cookies = http.cookies.SimpleCookie()
         self.set_status(status, reason)
 
@@ -407,7 +406,7 @@ class StreamResponse(HeadersMixin):
     def _copy_cookies(self):
         for cookie in self._cookies.values():
             value = cookie.output(header='')[1:]
-            self.headers.add('Set-Cookie', value)
+            self.headers.add(hdrs.SET_COOKIE, value)
 
     @property
     def started(self):
@@ -417,6 +416,14 @@ class StreamResponse(HeadersMixin):
     def status(self):
         return self._status
 
+    @property
+    def chunked(self):
+        return self._chunked
+
+    @property
+    def compression(self):
+        return self._compression
+
     @property
     def reason(self):
         return self._reason
@@ -434,6 +441,16 @@ class StreamResponse(HeadersMixin):
     def force_close(self):
         self._keep_alive = False
 
+    def enable_chunked_encoding(self, chunk_size=None):
+        """Enables automatic chunked transfer encoding."""
+        self._chunked = True
+        self._chunk_size = chunk_size
+
+    def enable_compression(self, force=False):
+        """Enables response compression with `deflate` encoding."""
+        self._compression = True
+        self._compression_force = force
+
     @property
     def headers(self):
         return self._headers
@@ -492,9 +509,9 @@ class StreamResponse(HeadersMixin):
         if value is not None:
             value = int(value)
             # TODO: raise error if chunked enabled
-            self.headers['Content-Length'] = str(value)
-        elif 'Content-Length' in self.headers:
-            del self.headers['Content-Length']
+            self.headers[hdrs.CONTENT_LENGTH] = str(value)
+        elif hdrs.CONTENT_LENGTH in self.headers:
+            del self.headers[hdrs.CONTENT_LENGTH]
 
     @property
     def content_type(self):
@@ -524,23 +541,30 @@ class StreamResponse(HeadersMixin):
             self._content_dict['charset'] = str(value).lower()
         self._generate_content_type_header()
 
-    def _generate_content_type_header(self):
+    def _generate_content_type_header(self, CONTENT_TYPE=hdrs.CONTENT_TYPE):
         params = '; '.join("%s=%s" % i for i in self._content_dict.items())
         if params:
             ctype = self._content_type + '; ' + params
         else:
             ctype = self._content_type
-        self.headers['Content-Type'] = ctype
+        self.headers[CONTENT_TYPE] = ctype
 
-    def start(self, request):
+    def _start_pre_check(self, request):
         if self._resp_impl is not None:
             if self._req is not request:
                 raise RuntimeError(
                     'Response has been started with different request.')
-            return self._resp_impl
+            else:
+                return self._resp_impl
+        else:
+            return None
+
+    def start(self, request):
+        resp_impl = self._start_pre_check(request)
+        if resp_impl is not None:
+            return resp_impl
 
         self._req = request
-
         keep_alive = self._keep_alive
         if keep_alive is None:
             keep_alive = request.keep_alive
@@ -555,7 +579,18 @@ class StreamResponse(HeadersMixin):
 
         self._copy_cookies()
 
-        headers = self.headers.items(getall=True)
+        if self._compression:
+            if (self._compression_force or
+                    'deflate' in request.headers.get(
+                        hdrs.ACCEPT_ENCODING, '')):
+                resp_impl.add_compression_filter()
+
+        if self._chunked:
+            resp_impl.enable_chunked_encoding()
+            if self._chunk_size:
+                resp_impl.add_chunking_filter(self._chunk_size)
+
+        headers = self.headers.items()
         for key, val in headers:
             resp_impl.add_header(key, val)
 
@@ -564,7 +599,7 @@ class StreamResponse(HeadersMixin):
 
     def write(self, data):
         if not isinstance(data, (bytes, bytearray, memoryview)):
-            raise TypeError('data argument must be byte-ish (%r)',
+            raise TypeError('data argument must be byte-ish (%r)' %
                             type(data))
 
         if self._eof_sent:
@@ -577,6 +612,12 @@ class StreamResponse(HeadersMixin):
         else:
             return ()
 
+    @asyncio.coroutine
+    def drain(self):
+        if self._resp_impl is None:
+            raise RuntimeError("Response has not been started")
+        yield from self._resp_impl.transport.drain()
+
     @asyncio.coroutine
     def write_eof(self):
         if self._eof_sent:
@@ -597,17 +638,32 @@ class Response(StreamResponse):
 
         if headers is not None:
             self.headers.extend(headers)
-        if content_type:
-            self.content_type = content_type
 
         if body is not None and text is not None:
             raise ValueError("body and text are not allowed together.")
-        elif body is not None:
-            self.body = body
-        elif text is not None:
-            self.text = text
+
+        if text is not None:
+            if hdrs.CONTENT_TYPE not in self.headers:
+                # fast path for filling headers
+                if not isinstance(text, str):
+                    raise TypeError('text argument must be str (%r)' %
+                                    type(text))
+                if content_type is None:
+                    content_type = 'text/plain'
+                self.headers[hdrs.CONTENT_TYPE] = (
+                    content_type + '; charset=utf-8')
+                self._content_type = content_type
+                self._content_dict = {'charset': 'utf-8'}
+                self.body = text.encode('utf-8')
+            else:
+                self.text = text
         else:
-            self.body = None
+            if content_type:
+                self.content_type = content_type
+            if body is not None:
+                self.body = body
+            else:
+                self.body = None
 
     @property
     def body(self):
@@ -616,7 +672,7 @@ class Response(StreamResponse):
     @body.setter
     def body(self, body):
         if body is not None and not isinstance(body, bytes):
-            raise TypeError('body argument must be bytes (%r)', type(body))
+            raise TypeError('body argument must be bytes (%r)' % type(body))
         self._body = body
         if body is not None:
             self.content_length = len(body)
@@ -630,7 +686,7 @@ class Response(StreamResponse):
     @text.setter
     def text(self, text):
         if text is not None and not isinstance(text, str):
-            raise TypeError('text argument must be str (%r)', type(text))
+            raise TypeError('text argument must be str (%r)' % type(text))
 
         if self.content_type == 'application/octet-stream':
             self.content_type = 'text/plain'
@@ -647,6 +703,171 @@ class Response(StreamResponse):
         yield from super().write_eof()
 
 
+class WebSocketResponse(StreamResponse):
+
+    def __init__(self, *, protocols=()):
+        super().__init__(status=101)
+        self._protocols = protocols
+        self._protocol = None
+        self._writer = None
+        self._reader = None
+        self._closing = False
+        self._loop = None
+        self._closing_fut = None
+
+    def start(self, request):
+        # make pre-check to don't hide it by do_handshake() exceptions
+        resp_impl = self._start_pre_check(request)
+        if resp_impl is not None:
+            return resp_impl
+
+        try:
+            status, headers, parser, writer, protocol = do_handshake(
+                request.method, request.headers, request.transport,
+                self._protocols)
+        except HttpProcessingError as err:
+            if err.code == 405:
+                raise HTTPMethodNotAllowed(request.method, ['GET'])
+            elif err.code == 400:
+                raise HTTPBadRequest(text=err.message, headers=err.headers)
+            else:  # pragma: no cover
+                raise HTTPInternalServerError() from err
+
+        if self.status != status:
+            self.set_status(status)
+        for k, v in headers:
+            self.headers[k] = v
+        self.force_close()
+
+        resp_impl = super().start(request)
+
+        self._reader = request._reader.set_parser(parser)
+        self._writer = writer
+        self._protocol = protocol
+        self._loop = request.app.loop
+        self._closing_fut = asyncio.Future(loop=self._loop)
+
+        return resp_impl
+
+    def can_start(self, request):
+        if self._writer is not None:
+            raise RuntimeError('Already started')
+        try:
+            _, _, _, _, protocol = do_handshake(
+                request.method, request.headers, request.transport,
+                self._protocols)
+        except HttpProcessingError:
+            return False, None
+        else:
+            return True, protocol
+
+    @property
+    def closing(self):
+        return self._closing
+
+    @property
+    def protocol(self):
+        return self._protocol
+
+    def ping(self, message='b'):
+        if self._writer is None:
+            raise RuntimeError('Call .start() first')
+        if self._closing:
+            raise RuntimeError('websocket connection is closing')
+        self._writer.ping(message)
+
+    def pong(self, message='b'):
+        # unsolicited pong
+        if self._writer is None:
+            raise RuntimeError('Call .start() first')
+        if self._closing:
+            raise RuntimeError('websocket connection is closing')
+        self._writer.pong(message)
+
+    def send_str(self, data):
+        if self._writer is None:
+            raise RuntimeError('Call .start() first')
+        if self._closing:
+            raise RuntimeError('websocket connection is closing')
+        if not isinstance(data, str):
+            raise TypeError('data argument must be str (%r)' % type(data))
+        self._writer.send(data, binary=False)
+
+    def send_bytes(self, data):
+        if self._writer is None:
+            raise RuntimeError('Call .start() first')
+        if self._closing:
+            raise RuntimeError('websocket connection is closing')
+        if not isinstance(data, (bytes, bytearray, memoryview)):
+            raise TypeError('data argument must be byte-ish (%r)' %
+                            type(data))
+        self._writer.send(data, binary=True)
+
+    def close(self, *, code=1000, message=b''):
+        if self._writer is None:
+            raise RuntimeError('Call .start() first')
+        if not self._closing:
+            self._closing = True
+            self._writer.close(code, message)
+        else:
+            raise RuntimeError('Already closing')
+
+    @asyncio.coroutine
+    def wait_closed(self):
+        if self._closing_fut is None:
+            raise RuntimeError('Call .start() first')
+        yield from self._closing_fut
+
+    @asyncio.coroutine
+    def receive_msg(self):
+        if self._reader is None:
+            raise RuntimeError('Call .start() first')
+        while True:
+            try:
+                msg = yield from self._reader.read()
+            except Exception as exc:
+                self._closing_fut.set_exception(exc)
+                raise
+
+            if msg.tp == MSG_CLOSE:
+                if self._closing:
+                    exc = WSClientDisconnectedError(msg.data, msg.extra)
+                    self._closing_fut.set_exception(exc)
+                    raise exc
+                else:
+                    self._closing = True
+                    self._writer.close(msg.data, msg.extra)
+                    yield from self.drain()
+                    exc = WSClientDisconnectedError(msg.data, msg.extra)
+                    self._closing_fut.set_exception(exc)
+                    raise exc
+            elif not self._closing:
+                if msg.tp == MSG_PING:
+                    self._writer.pong(msg.data)
+                elif msg.tp in (MSG_TEXT, MSG_BINARY):
+                    return msg
+
+    @asyncio.coroutine
+    def receive_str(self):
+        msg = yield from self.receive_msg()
+        if msg.tp != MSG_TEXT:
+            raise TypeError(
+                "Received message {}:{!r} is not str".format(msg.tp, msg.data))
+        return msg.data
+
+    @asyncio.coroutine
+    def receive_bytes(self):
+        msg = yield from self.receive_msg()
+        if msg.tp != MSG_BINARY:
+            raise TypeError(
+                "Received message {}:{!r} is not bytes".format(msg.tp,
+                                                               msg.data))
+        return msg.data
+
+    def write(self, data):
+        raise RuntimeError("Cannot call .write() for websocket")
+
+
 ############################################################
 # HTTP Exceptions
 ############################################################
@@ -950,7 +1171,7 @@ class PlainRoute(Route):
         self._path = path
 
     def match(self, path):
-        # string comparsion about 10 times faster than regexp matching
+        # string comparsion is about 10 times faster than regexp matching
         if self._path == path:
             return {}
         else:
@@ -993,6 +1214,8 @@ class DynamicRoute(Route):
 
 class StaticRoute(Route):
 
+    limit = 8192
+
     def __init__(self, name, prefix, directory):
         assert prefix.startswith('/'), prefix
         assert prefix.endswith('/'), prefix
@@ -1027,16 +1250,22 @@ class StaticRoute(Route):
             ct = 'application/octet-stream'
         resp.content_type = ct
 
-        resp.headers['transfer-encoding'] = 'chunked'
+        file_size = os.stat(filepath).st_size
+        single_chunk = file_size < self.limit
+
+        if single_chunk:
+            resp.content_length = file_size
         resp.start(request)
 
         with open(filepath, 'rb') as f:
-            chunk = f.read(8192)
-            while chunk:
+            chunk = f.read(self.limit)
+            if single_chunk:
                 resp.write(chunk)
-                chunk = f.read(8192)
+            else:
+                while chunk:
+                    resp.write(chunk)
+                    chunk = f.read(self.limit)
 
-        yield from resp.write_eof()
         return resp
 
     def __repr__(self):
@@ -1046,6 +1275,44 @@ class StaticRoute(Route):
             directory=self._directory)
 
 
+class _NotFoundMatchInfo(UrlMappingMatchInfo):
+
+    def __init__(self):
+        super().__init__({}, None)
+
+    @property
+    def handler(self):
+        return self._not_found
+
+    @property
+    def route(self):
+        return None
+
+    @asyncio.coroutine
+    def _not_found(self, request):
+        raise HTTPNotFound()
+
+
+class _MethodNotAllowedMatchInfo(UrlMappingMatchInfo):
+
+    def __init__(self, method, allowed_methods):
+        super().__init__({}, None)
+        self._method = method
+        self._allowed_methods = allowed_methods
+
+    @property
+    def handler(self):
+        return self._not_allowed
+
+    @property
+    def route(self):
+        return None
+
+    @asyncio.coroutine
+    def _not_allowed(self, request):
+        raise HTTPMethodNotAllowed(self._method, self._allowed_methods)
+
+
 class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
 
     DYN = re.compile(r'^\{(?P<var>[a-zA-Z][_a-zA-Z0-9]*)\}$')
@@ -1077,9 +1344,9 @@ class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
                 return UrlMappingMatchInfo(match_dict, route)
         else:
             if allowed_methods:
-                raise HTTPMethodNotAllowed(method, allowed_methods)
+                return _MethodNotAllowedMatchInfo(method, allowed_methods)
             else:
-                raise HTTPNotFound()
+                return _NotFoundMatchInfo()
 
     def __iter__(self):
         return iter(self._routes)
@@ -1195,7 +1462,7 @@ class RequestHandler(ServerHttpProtocol):
 
         app = self._app
         request = Request(app, message, payload,
-                          self.transport, self.writer, self.keep_alive_timeout)
+                          self.transport, self.reader, self.writer)
         try:
             match_info = yield from self._router.resolve(request)
 
@@ -1210,9 +1477,11 @@ class RequestHandler(ServerHttpProtocol):
 
             if not isinstance(resp, StreamResponse):
                 raise RuntimeError(
-                    ("Handler {!r} should return response instance, got {!r} "
-                     "[middlewares {!r}]")
-                    .format(match_info.handler, type(resp), self._middlewares))
+                    ("Handler {!r} should return response instance, "
+                     "got {!r} [middlewares {!r}]").format(
+                         match_info.handler,
+                         type(resp),
+                         self._middlewares))
         except HTTPException as exc:
             resp = exc
 
@@ -1284,7 +1553,7 @@ class Application(dict):
 
     def __init__(self, *, logger=web_logger, loop=None,
                  router=None, handler_factory=RequestHandlerFactory,
-                 middlewares=(), **kwargs):
+                 middlewares=()):
         if loop is None:
             loop = asyncio.get_event_loop()
         if router is None:
@@ -1297,7 +1566,6 @@ class Application(dict):
         self._loop = loop
         self.logger = logger
 
-        self.update(**kwargs)
         for factory in middlewares:
             assert asyncio.iscoroutinefunction(factory), factory
         self._middlewares = tuple(middlewares)
