@@ -2,7 +2,7 @@ import asyncio
 import unittest
 from unittest import mock
 from aiohttp.web import Request
-from aiohttp.multidict import MultiDict
+from aiohttp.multidict import MultiDict, CIMultiDict
 from aiohttp.protocol import HttpVersion
 from aiohttp.protocol import RawRequestMessage
 
@@ -16,17 +16,17 @@ class TestWebRequest(unittest.TestCase):
     def tearDown(self):
         self.loop.close()
 
-    def make_request(self, method, path, headers=MultiDict(), *,
-                     version=HttpVersion(1, 1), closing=False,
-                     keep_alive_timeout=15):
+    def make_request(self, method, path, headers=CIMultiDict(), *,
+                     version=HttpVersion(1, 1), closing=False):
         self.app = mock.Mock()
         message = RawRequestMessage(method, path, version, headers, closing,
                                     False)
         self.payload = mock.Mock()
         self.transport = mock.Mock()
         self.writer = mock.Mock()
+        self.reader = mock.Mock()
         req = Request(self.app, message, self.payload,
-                      self.transport, self.writer, keep_alive_timeout)
+                      self.transport, self.reader, self.writer)
         return req
 
     def test_ctor(self):
@@ -66,20 +66,20 @@ class TestWebRequest(unittest.TestCase):
     def test_content_type_from_spec(self):
         req = self.make_request(
             'Get', '/',
-            MultiDict([('CONTENT-TYPE', 'application/json')]))
+            CIMultiDict([('CONTENT-TYPE', 'application/json')]))
         self.assertEqual('application/json', req.content_type)
 
     def test_content_type_from_spec_with_charset(self):
         req = self.make_request(
             'Get', '/',
-            MultiDict([('CONTENT-TYPE', 'text/html; charset=UTF-8')]))
+            CIMultiDict([('CONTENT-TYPE', 'text/html; charset=UTF-8')]))
         self.assertEqual('text/html', req.content_type)
         self.assertEqual('UTF-8', req.charset)
 
     def test_calc_content_type_on_getting_charset(self):
         req = self.make_request(
             'Get', '/',
-            MultiDict([('CONTENT-TYPE', 'text/html; charset=UTF-8')]))
+            CIMultiDict([('CONTENT-TYPE', 'text/html; charset=UTF-8')]))
         self.assertEqual('UTF-8', req.charset)
         self.assertEqual('text/html', req.content_type)
 
@@ -94,8 +94,9 @@ class TestWebRequest(unittest.TestCase):
         self.assertEqual('/путь', req.path)
 
     def test_content_length(self):
-        req = self.make_request('Get', '/',
-                                MultiDict([('CONTENT-LENGTH', '123')]))
+        req = self.make_request(
+            'Get', '/',
+            CIMultiDict([('CONTENT-LENGTH', '123')]))
 
         self.assertEqual(123, req.content_length)
 
@@ -107,23 +108,19 @@ class TestWebRequest(unittest.TestCase):
         req = self.make_request('GET', '/', closing=True)
         self.assertFalse(req.keep_alive)
 
-    def test_non_keepalive_on_server_protocol_without_keepalive_timeout(self):
-        req = self.make_request('GET', '/', closing=True, keep_alive_timeout=0)
-        self.assertFalse(req.keep_alive)
-
     def test_call_POST_on_GET_request(self):
         req = self.make_request('GET', '/')
 
         ret = self.loop.run_until_complete(req.post())
-        self.assertEqual(MultiDict(), ret)
+        self.assertEqual(CIMultiDict(), ret)
 
     def test_call_POST_on_weird_content_type(self):
         req = self.make_request(
             'POST', '/',
-            headers=MultiDict({'CONTENT-TYPE': 'something/weird'}))
+            headers=CIMultiDict({'CONTENT-TYPE': 'something/weird'}))
 
         ret = self.loop.run_until_complete(req.post())
-        self.assertEqual(MultiDict(), ret)
+        self.assertEqual(CIMultiDict(), ret)
 
     def test_call_POST_twice(self):
         req = self.make_request('GET', '/')
@@ -141,7 +138,7 @@ class TestWebRequest(unittest.TestCase):
         self.assertIs(cookies, req.cookies)
 
     def test_request_cookie(self):
-        headers = MultiDict(COOKIE='cookie1=value1; cookie2=value2')
+        headers = CIMultiDict(COOKIE='cookie1=value1; cookie2=value2')
         req = self.make_request('GET', '/', headers=headers)
 
         self.assertEqual(req.cookies, {
@@ -150,7 +147,7 @@ class TestWebRequest(unittest.TestCase):
         })
 
     def test_request_cookie__set_item(self):
-        headers = MultiDict(COOKIE='name=value')
+        headers = CIMultiDict(COOKIE='name=value')
         req = self.make_request('GET', '/', headers=headers)
 
         self.assertEqual(req.cookies, {'name': 'value'})
@@ -164,3 +161,9 @@ class TestWebRequest(unittest.TestCase):
         match = {'a': 'b'}
         req._match_info = match
         self.assertIs(match, req.match_info)
+
+    def test_request_is_dict(self):
+        req = self.make_request('GET', '/')
+        self.assertTrue(isinstance(req, dict))
+        req['key'] = 'value'
+        self.assertEqual('value', req['key'])
