@@ -1,7 +1,8 @@
 import asyncio
 import unittest
 from unittest import mock
-from aiohttp.multidict import MultiDict, CaseInsensitiveMultiDict
+from aiohttp import hdrs
+from aiohttp.multidict import CIMultiDict
 from aiohttp.web import Request, StreamResponse, Response
 from aiohttp.protocol import RawRequestMessage, HttpVersion11
 
@@ -15,15 +16,16 @@ class TestStreamResponse(unittest.TestCase):
     def tearDown(self):
         self.loop.close()
 
-    def make_request(self, method, path, headers=MultiDict()):
+    def make_request(self, method, path, headers=CIMultiDict()):
         self.app = mock.Mock()
         message = RawRequestMessage(method, path, HttpVersion11, headers,
                                     False, False)
         self.payload = mock.Mock()
         self.transport = mock.Mock()
+        self.reader = mock.Mock()
         self.writer = mock.Mock()
         req = Request(self.app, message, self.payload,
-                      self.transport, self.writer, 15)
+                      self.transport, self.reader, self.writer)
         return req
 
     def test_ctor(self):
@@ -114,6 +116,75 @@ class TestStreamResponse(unittest.TestCase):
         with self.assertRaises(RuntimeError):
             resp.start(req2)
 
+    @mock.patch('aiohttp.web.ResponseImpl')
+    def test_chunked_encoding(self, ResponseImpl):
+        req = self.make_request('GET', '/')
+        resp = StreamResponse()
+        self.assertFalse(resp.chunked)
+
+        resp.enable_chunked_encoding()
+        self.assertTrue(resp.chunked)
+
+        msg = resp.start(req)
+        self.assertTrue(msg.chunked)
+
+    @mock.patch('aiohttp.web.ResponseImpl')
+    def test_chunk_size(self, ResponseImpl):
+        req = self.make_request('GET', '/')
+        resp = StreamResponse()
+        self.assertFalse(resp.chunked)
+
+        resp.enable_chunked_encoding(chunk_size=8192)
+        self.assertTrue(resp.chunked)
+
+        msg = resp.start(req)
+        self.assertTrue(msg.chunked)
+        msg.add_chunking_filter.assert_called_with(8192)
+        self.assertIsNotNone(msg.filter)
+
+    @mock.patch('aiohttp.web.ResponseImpl')
+    def test_compression_no_accept(self, ResponseImpl):
+        req = self.make_request('GET', '/')
+        resp = StreamResponse()
+        self.assertFalse(resp.chunked)
+
+        self.assertFalse(resp.compression)
+        resp.enable_compression()
+        self.assertTrue(resp.compression)
+
+        msg = resp.start(req)
+        self.assertFalse(msg.add_compression_filter.called)
+
+    @mock.patch('aiohttp.web.ResponseImpl')
+    def test_force_compression_no_accept(self, ResponseImpl):
+        req = self.make_request('GET', '/')
+        resp = StreamResponse()
+        self.assertFalse(resp.chunked)
+
+        self.assertFalse(resp.compression)
+        resp.enable_compression(force=True)
+        self.assertTrue(resp.compression)
+
+        msg = resp.start(req)
+        self.assertTrue(msg.add_compression_filter.called)
+        self.assertIsNotNone(msg.filter)
+
+    @mock.patch('aiohttp.web.ResponseImpl')
+    def test_compression(self, ResponseImpl):
+        req = self.make_request(
+            'GET', '/',
+            headers=CIMultiDict({hdrs.ACCEPT_ENCODING: 'gzip, deflate'}))
+        resp = StreamResponse()
+        self.assertFalse(resp.chunked)
+
+        self.assertFalse(resp.compression)
+        resp.enable_compression()
+        self.assertTrue(resp.compression)
+
+        msg = resp.start(req)
+        self.assertTrue(msg.add_compression_filter.called)
+        self.assertIsNotNone(msg.filter)
+
     def test_write_non_byteish(self):
         resp = StreamResponse()
         resp.start(self.make_request('GET', '/'))
@@ -270,15 +341,16 @@ class TestResponse(unittest.TestCase):
     def tearDown(self):
         self.loop.close()
 
-    def make_request(self, method, path, headers=MultiDict()):
+    def make_request(self, method, path, headers=CIMultiDict()):
         self.app = mock.Mock()
         message = RawRequestMessage(method, path, HttpVersion11, headers,
                                     False, False)
         self.payload = mock.Mock()
         self.transport = mock.Mock()
+        self.reader = mock.Mock()
         self.writer = mock.Mock()
         req = Request(self.app, message, self.payload,
-                      self.transport, self.writer, 15)
+                      self.transport, self.reader, self.writer)
         return req
 
     def test_ctor(self):
@@ -288,7 +360,7 @@ class TestResponse(unittest.TestCase):
         self.assertEqual('OK', resp.reason)
         self.assertIsNone(resp.body)
         self.assertEqual(0, resp.content_length)
-        self.assertEqual(CaseInsensitiveMultiDict([('CONTENT-LENGTH', '0')]),
+        self.assertEqual(CIMultiDict([('CONTENT-LENGTH', '0')]),
                          resp.headers)
 
     def test_ctor_with_headers_and_status(self):
@@ -297,9 +369,9 @@ class TestResponse(unittest.TestCase):
         self.assertEqual(201, resp.status)
         self.assertEqual(b'body', resp.body)
         self.assertEqual(4, resp.content_length)
-        self.assertEqual(CaseInsensitiveMultiDict([('AGE', '12'),
-                                                   ('CONTENT-LENGTH', '4')]),
-                         resp.headers)
+        self.assertEqual(CIMultiDict(
+            [('AGE', '12'),
+             ('CONTENT-LENGTH', '4')]), resp.headers)
 
     def test_ctor_content_type(self):
         resp = Response(content_type='application/json')
@@ -307,8 +379,9 @@ class TestResponse(unittest.TestCase):
         self.assertEqual(200, resp.status)
         self.assertEqual('OK', resp.reason)
         self.assertEqual(
-            CaseInsensitiveMultiDict([('CONTENT-TYPE', 'application/json'),
-                                      ('CONTENT-LENGTH', '0')]),
+            CIMultiDict(
+                [('CONTENT-TYPE', 'application/json'),
+                 ('CONTENT-LENGTH', '0')]),
             resp.headers)
 
     def test_ctor_text_body_combined(self):
@@ -321,7 +394,7 @@ class TestResponse(unittest.TestCase):
         self.assertEqual(200, resp.status)
         self.assertEqual('OK', resp.reason)
         self.assertEqual(
-            CaseInsensitiveMultiDict(
+            CIMultiDict(
                 [('CONTENT-TYPE', 'text/plain; charset=utf-8'),
                  ('CONTENT-LENGTH', '9')]),
             resp.headers)
@@ -434,3 +507,29 @@ class TestResponse(unittest.TestCase):
         resp = StreamResponse()
         resp.start(self.make_request('GET', '/'))
         self.assertTrue(resp.started)
+
+    def test_drain_before_start(self):
+
+        @asyncio.coroutine
+        def go():
+            resp = StreamResponse()
+            with self.assertRaises(RuntimeError):
+                yield from resp.drain()
+
+        self.loop.run_until_complete(go())
+
+    def test_nonstr_text_in_ctor(self):
+        with self.assertRaises(TypeError):
+            Response(text=b'data')
+
+    def test_text_in_ctor_with_content_type(self):
+        resp = Response(text='data', content_type='text/html')
+        self.assertEqual('data', resp.text)
+        self.assertEqual('text/html', resp.content_type)
+
+    def test_text_in_ctor_with_content_type_header(self):
+        resp = Response(text='текст',
+                        headers={'Content-Type': 'text/html; charset=koi8-r'})
+        self.assertEqual('текст'.encode('koi8-r'), resp.body)
+        self.assertEqual('text/html', resp.content_type)
+        self.assertEqual('koi8-r', resp.charset)
